import queue
from typing import List, Tuple, Set
import logging
import time

from asciimatics.widgets import Widget, Frame
from asciimatics.constants import (
    COLOUR_GREEN, COLOUR_CYAN, COLOUR_BLACK, COLOUR_WHITE, COLOUR_BLUE, COLOUR_RED, COLOUR_YELLOW, COLOUR_MAGENTA,
    A_NORMAL, A_BOLD, A_UNDERLINE, A_REVERSE
)
from asciimatics.widgets.utilities import _split_text

from j_chess_client_manager.logging import LOG_QUEUE


_LOG_LIST: List[Tuple[str, logging.LogRecord]] = []


def _update_log_list():
    while not LOG_QUEUE.empty():
        try:
            _LOG_LIST.append(LOG_QUEUE.get())
        except queue.Empty:
            break

def _get_logs(*codes: str) -> List[Tuple[str, logging.LogRecord]]:
    if codes is None or len(codes) <= 0:
        return _LOG_LIST
    return [x for x in _LOG_LIST if x[0] in codes]

def _get_codes() -> Set[str]:
    return {x[0] for x in _LOG_LIST}


class LogList(Widget):

    def __init__(self, name, height: int = 15):
        super().__init__(name, tab_stop=False)
        self._height = height
        self._fmt = '{time} [{code:^4s}-{levelname:^8s}] - {message}'
        self.date_fmt = "%Y-%m-%d %H:%M:%S"

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, value: int):
        self._height = value

    def update(self, frame_no):
        _update_log_list()
        frame: Frame = self.frame
        print_logs = _get_logs()
        (colour, attr, background) = frame.palette["label"]
        frame.canvas.paint(f"Logs [{len(print_logs)}]", self._x, self._y, colour, A_UNDERLINE, background)
        print_records = print_logs[-(self._height - 1):]
        log_offset = 4

        for i, (code, record) in enumerate(print_records):
            record: logging.LogRecord
            t = time.strftime(self.date_fmt, time.localtime(record.created))
            line = _split_text(self._fmt.format(time=t, code=code, levelname=record.levelname, message=record.message),
                               width=self.width - log_offset, height=1, unicode_aware=True)[0]
            frame.canvas.paint(line, self._x + log_offset, self._y + i + 1, colour, A_NORMAL, background)

    def reset(self):
        pass

    def process_event(self, event):
        return event

    def required_height(self, offset, width):
        return self.height

    def value(self):
        return self._logs



"""Top-level package for j-chess client manager."""
from j_chess_client_manager.logging import SYSTEM_LOGGER
from j_chess_lib import __version__ as __lib__version__, __schema_version__, __author__ as __lib__author__

__author__ = """RedRem95"""
__email__ = 'redrem@botschmot.de'
__version__ = '0.1.5'
__project_name__ = "J-Chess Client Manager"

SYSTEM_LOGGER.info(f"Running \"{__project_name__}\" Version {__version__} by {__author__}")
SYSTEM_LOGGER.info(f"Using \"j_chess_lib\" Version {__lib__version__} by {__lib__author__} "
                   f"with schema version {__schema_version__}")



from abc import ABC, abstractmethod
from typing import List, Tuple


class Piece(ABC):

    @classmethod
    @abstractmethod
    def _paths(cls) -> List[List[Tuple[float, float]]]:
        pass

    @classmethod
    def path(cls, w: int, h: int, dx: int = 0, dy: int = 0) -> List[List[Tuple[int, int]]]:
        return [
            [(int(w_f * w) + dx, int(h_f * h) + dy) for w_f, h_f in x] for x in cls._paths()
        ]


class Pawn(Piece):

    @classmethod
    def _paths(cls) -> List[List[Tuple[float, float]]]:
        return [
            [(0.33, 0.9), (0.66, 0.9), (0.66, 0.2), (0.33, 0.2)]
        ]



from inspect import Parameter, _empty, isclass
from typing import Callable, Any, Tuple, List, Type, Set

from asciimatics.widgets import Text, Widget
from j_chess_lib.ai import AI


def all_subclasses(cls) -> Set[Type[AI]]:
    return set(cls.__subclasses__()).union([s for c in cls.__subclasses__() for s in all_subclasses(c)])


def get_all_ais() -> List[Type[AI]]:
    return sorted([x for x in all_subclasses(AI) if len(x.__abstractmethods__) == 0 and not x.__name__.startswith("_")],
                  key=lambda x: x.__name__)


def get_widget_by_parameter(
    name_base: str, parameter: Parameter, none_const: str = "<<None>>"
) -> Tuple[Widget, Callable[[Any], Any]]:
    name = parameter.name
    default = none_const if parameter.default is _empty or parameter.default is None else str(parameter.default)
    annotation: Any = parameter.annotation

    def annotation_convert(val: str):
        if val == none_const:
            return None
        return val

    if annotation is _empty:
        annotation = None
    else:
        if isclass(annotation):
            annotation_class = annotation

            def annotation_convert(val):
                if val == none_const:
                    return None
                # noinspection PyCallingNonCallable
                return annotation_class(val)

            annotation = annotation.__name__
        else:
            annotation = str(annotation)

    def validator(val: str):
        try:
            annotation_convert(val=val)
        except Exception as e:
            return False
        return True

    widget = Text(label=f"{name}{'' if annotation is None else f' [{annotation}]'}",
                  name=f"{name_base}__{name}", validator=validator)
    widget.value = default

    return widget, annotation_convert



from typing import List, Tuple, Union, Optional

from asciimatics.constants import (
    COLOUR_GREEN, COLOUR_CYAN, COLOUR_BLACK, COLOUR_WHITE, COLOUR_BLUE, COLOUR_RED, COLOUR_YELLOW, COLOUR_MAGENTA,
    A_NORMAL, A_BOLD, A_UNDERLINE, A_REVERSE
)
from asciimatics.widgets import Widget, Frame

# noinspection PyProtectedMember
from j_chess_client_manager.clients import SuperProvider, _NoneProvider
from j_chess_client_manager.logging import SYSTEM_LOGGER

_none_provider = _NoneProvider()


class ChessBoard(Widget):

    def __init__(self, w_h_factor: float = 2):
        super().__init__("Chessboard", tab_stop=False)
        self.disabled = True
        self._w_h_factor = w_h_factor
        self._provider: Optional[SuperProvider] = None
        self._empty_board: List[List[str]] = [["" for _ in range(8)] for _ in range(8)]

    def _draw_square(self, frame: Frame, x: int, y: int, w: int, h: int, colour: int, bg: int):
        self._draw_path(frame=frame, path=[[(x, y), (x + w, y), (x + w, y + h), (x, y + h)]], colour=colour, bg=bg)

    def _draw_path(self, frame: Frame, path: List[List[Tuple[int, int]]], colour: int, bg: int):
        _x = self._x
        _y = self._y
        path = [[(x + _x, y + _y) for x, y in p] for p in path]
        frame.canvas.fill_polygon(path, colour=colour, bg=bg)

    @property
    def data_provider(self) -> Optional[SuperProvider]:
        if self._provider is None:
            return _none_provider
        return self._provider

    @data_provider.setter
    def data_provider(self, value: SuperProvider):
        self._provider = value

    def update(self, frame_no):
        frame = self.frame
        status_height = self._draw_status(frame=frame, margin=3)
        self._draw_chess_board(frame=frame, y_off=status_height)

    def _draw_status(self, frame: Frame, margin: int = 0) -> int:
        (colour, attr, background) = frame.palette["label"]
        attr_active_player = A_BOLD
        white_text = f"{self.data_provider.white_name} {self.data_provider.white_time}s"
        black_text = f"{self.data_provider.black_name} {self.data_provider.black_time}s"
        mid_text = f"Turn: {self.data_provider.turn()}"
        frame.canvas.paint(
            white_text, self._x + margin, self._y,
            colour, attr_active_player if self.data_provider.white_turn() else A_NORMAL, background
        )
        frame.canvas.paint(
            black_text, self._x + self.width - margin - len(black_text), self._y,
            colour, attr_active_player if not self.data_provider.white_turn() else A_NORMAL, background
        )
        frame.canvas.paint(
            mid_text, self._x + (self.width // 2) - (len(mid_text) // 2), self._y,
            colour, attr_active_player, background
        )
        return 1

    def _draw_chess_board(self, frame: Frame, colour_black: int = COLOUR_BLUE, colour_white: int = COLOUR_CYAN,
                          y_off: int = 0):

        current_board = None if self.data_provider is None else self.data_provider.board()
        if current_board is None:
            current_board = self._empty_board

        w, h = self.width, self._h - y_off

        board_h = h
        board_w = board_h * self._w_h_factor

        if board_w > w:
            board_w = w
            board_h = board_w // 2

        dx = int(1 * self._w_h_factor)
        dy = 1
        if dx < 1:
            dx = 1
            dy = int(1 / self._w_h_factor)

        tile_w = (board_w - 2 * dx) // 8
        tile_h = (board_h - 2 * dy) // 8

        board_w = tile_w * 8
        board_h = tile_h * 8

        offset_x = (w - board_w) // 2
        offset_y = (h - board_h) // 2 + y_off

        self._draw_square(frame=frame, x=offset_x - dx, y=offset_y - dy, w=board_w + 2 * dx, h=board_h + 2 * dy,
                          colour=COLOUR_GREEN, bg=COLOUR_BLACK)

        for x in range(8):
            for y in range(8):
                tile_colour = colour_white if (x + y) % 2 == 0 else colour_black
                _x, _y = offset_x + (x * tile_w), offset_y + y * tile_h
                self._draw_square(frame=frame, x=_x, y=_y, w=tile_w, h=tile_h, colour=tile_colour, bg=COLOUR_GREEN)
                frame.canvas.paint(
                    f"{chr(97 + x)}{8-y}", self._x + _x, self._y + _y, COLOUR_RED, A_BOLD, tile_colour
                )
                piece: str = current_board[y][x]
                if len(piece) == 1:
                    frame.canvas.paint(
                        piece.upper(), self._x + _x + tile_w//2, self._y + _y + tile_h//2,
                        COLOUR_BLACK if piece.islower() else COLOUR_WHITE, A_BOLD, tile_colour
                    )

        frame.canvas.paint(
            "Board: {}x{}; Tile: {}x{}".format(board_w, board_h, tile_w, tile_h),
            self._x + offset_x - dx, self._y + offset_y - dy, COLOUR_BLACK, A_NORMAL, COLOUR_GREEN
        )

    def reset(self):
        pass

    def process_event(self, event):
        return event

    def required_height(self, offset, width):
        return Widget.FILL_FRAME

    def value(self):
        return None



from typing import Callable, Any, List, Optional

from asciimatics.exceptions import StopApplication, ResizeScreenError
from asciimatics.scene import Scene
from asciimatics.screen import Screen
from asciimatics.effects import Print, BannerText
from asciimatics.renderers import StaticRenderer, ColourImageFile

from j_chess_client_manager.ui.frames.main_frame import MainFrame
from j_chess_client_manager.ui.frames.client_view import ClientView


def setup_scenes(screen: Screen, theme: str) -> List[Scene]:
    scenes = []

    mf = MainFrame(screen=screen, theme=theme)

    scenes.append(
        Scene([mf], -1, name="Main")
    )
    scenes.append(
        Scene([ClientView(screen=screen, ai_adder=mf.add_ai, theme=theme)], -1, name="Add Client")
    )

    return scenes


def run_function_creator(start_scene: Any, theme: str = "default") -> Callable[[Screen], None]:

    def run(screen: Screen):
        scenes = setup_scenes(screen=screen, theme=theme)

        screen.play(scenes, stop_on_resize=True, repeat=False, start_scene=start_scene)

    return run



import logging
from logging import LogRecord
from logging.handlers import QueueHandler
import queue
from typing import Tuple

from j_chess_lib import logger as _lib_logger

_lib_logger.handlers = []


class _CustomQueueHandler(QueueHandler):

    def __init__(self, code: str, log_queue):
        super().__init__(log_queue)
        self._code = code

    def enqueue(self, record: LogRecord) -> None:
        self.queue.put((self._code, record))


LOG_QUEUE = queue.Queue()


def get_log_handler(code: str) -> logging.Handler:
    """
    Return a handler pushing to the log queue so it can be displayed in the UI

    Parameters
    ----------
    code: str
        Symbols to identify this handler in the UI. Should be at max 4 symbols

    Returns
    -------
    New handler
    """
    new_handler = _CustomQueueHandler(code=code, log_queue=LOG_QUEUE)
    new_handler.setLevel(logging.INFO)
    return new_handler


_lib_logger.addHandler(get_log_handler("LIB"))
_lib_logger.setLevel(logging.DEBUG)

UI_LOGGER = logging.getLogger("Manager-UI")
UI_LOGGER.addHandler(get_log_handler("UI"))
UI_LOGGER.setLevel(logging.DEBUG)

SYSTEM_LOGGER = logging.getLogger("Backend-System")
SYSTEM_LOGGER.addHandler(get_log_handler("SYS"))
SYSTEM_LOGGER.setLevel(logging.DEBUG)




from abc import ABC, abstractmethod
from enum import Enum
from typing import Type, Optional, List, Tuple

from j_chess_lib.ai import AI

from .metrics import MetricProvider
from .board import BoardProvider
from .refresh import RefreshInitiator


class ClientTypes(Enum):
    AI = 0
    Spectator = 1
    Unknown = None

    def __str__(self, question_tag_len: int = 1):
        if self == self.Unknown:
            return "?" * question_tag_len
        if self == self.Spectator:
            return "Spec"
        return self.name

    @classmethod
    def max_len(cls):
        return max(len(str(x)) for x in cls)

    def fixed_represent(self):
        max_len = self.max_len()
        return f"[{self.__str__(question_tag_len=max_len):^{max_len}s}]"

    def get_class(self) -> Optional[Type]:
        if self == self.AI:
            return AI
        return None


class Typeable:

    def get_client_type(self) -> ClientTypes:
        for possible_type in ClientTypes:
            possible_type: ClientTypes
            needed_class = possible_type.get_class()
            if needed_class is not None and isinstance(self, needed_class):
                return possible_type
        return ClientTypes.Unknown


class SuperProvider(MetricProvider, BoardProvider, Typeable, ABC):
    pass


class _NoneProvider(SuperProvider):
    def metrics(self) -> List[Tuple[Tuple[str, str], int]]:
        return []

    @property
    def fen(self) -> Optional[str]:
        return None

    @property
    def white_name(self):
        return "---"

    @property
    def black_name(self):
        return "---"

    @property
    def white_time(self):
        return -1

    @property
    def black_time(self):
        return -1

    @property
    def need_refresh(self) -> bool:
        return False



from abc import ABC, abstractmethod
from typing import List, Tuple, Dict, Optional


class BoardProvider(ABC):

    @property
    @abstractmethod
    def fen(self) -> Optional[str]:
        pass

    # rnbqkbnr/pp1ppppp/8/2p5/4P3/5N2/PPPP1PPP/RNBQKB1R b KQkq - 1 2

    def board(self) -> List[List[str]]:
        fen = self.fen
        if fen is None or len(fen) <= 0:
            return [["" for _ in range(8)] for _ in range(8)]
        ret: List[List[str]] = [[]]
        board = self.fen.split(" ")[0]
        for c in board:
            try:
                n = int(c)
                for _ in range(n):
                    ret[-1].append("")
            except ValueError:
                if c == "/":
                    ret.append([])
                else:
                    ret[-1].append(c)

        return ret

    def turn(self) -> int:
        try:
            return int(self.fen.split(" ")[-1])
        except (ValueError, AttributeError, KeyError):
            return -1

    def white_turn(self) -> bool:
        try:
            return self.fen.split(" ")[1].lower() == "w"
        except (ValueError, AttributeError, KeyError):
            return True

    def castling(self) -> Dict[str, Dict[str, bool]]:
        try:
            castling = self.fen.split(" ")[2]
            return {
                "w": {"k": "K" in castling, "q": "Q" in castling},
                "b": {"k": "k" in castling, "q": "q" in castling}
            }
        except (ValueError, AttributeError, KeyError):
            return {
                "w": {"k": False, "q": False},
                "b": {"k": False, "q": False}
            }

    def en_passant(self) -> Optional[str]:
        try:
            en_passant = self.fen.split(" ")[3]
            return None if en_passant == "-" else en_passant.strip()
        except (ValueError, AttributeError, KeyError):
            return ""

    def half_moves_since_pawn(self) -> int:
        try:
            return int(self.fen.split(" ")[4])
        except (ValueError, AttributeError, KeyError):
            return -1

    @property
    @abstractmethod
    def white_name(self):
        pass

    @property
    @abstractmethod
    def black_name(self):
        pass

    @property
    @abstractmethod
    def white_time(self):
        pass

    @property
    @abstractmethod
    def black_time(self):
        pass



from abc import ABC
from typing import Type, Any, Dict, Optional, Tuple, List, Union, Callable
from uuid import UUID

from j_chess_lib.ai import StoreAI, AI
from j_chess_lib.ai.Container import GameState
from j_chess_lib.communication import MoveData, MatchStatusData, MatchFormatData

from . import SuperProvider, MetricProvider
from j_chess_client_manager.logging import SYSTEM_LOGGER
from j_chess_lib.ai.Sample import SampleAI


def wrap_ai(
    base_ai: Type[AI], init_values: Dict[str, Any], need_update: Callable[[Union[AI, SuperProvider]], None] = None
) -> Union[AI, SuperProvider]:
    if need_update is None:
        def need_update(*args, **kwargs):
            pass

    class _WrappedAI(SuperProvider, base_ai):

        def __init__(self, base_init_values: Dict[str, Any]):
            super(_WrappedAI, self).__init__(**base_init_values)
            self._need_refresh: bool = False
            self._fen = None
            self._enemy_name = None
            self._i_am_white = True
            self._your_time = -1
            self._enemy_time = -1

        @property
        def white_time(self) -> int:
            return self._your_time if self._i_am_white else self._enemy_time

        @property
        def black_time(self) -> int:
            return self._your_time if not self._i_am_white else self._enemy_time

        @property
        def fen(self) -> str:
            return self._fen

        def metrics(self) -> List[Tuple[Tuple[str, str], int]]:
            en_passant = self.en_passant()
            castling = self.castling()
            ret = [
                (("Turn", str(self.turn())), 1),
                (("Current player", "white" if self.white_turn() else "black"), 2),
                (("En passant", "---" if en_passant is None else en_passant), 3),
                (("Castling white", f"King: {castling['w']['k']}; Queen: {castling['w']['q']}"), 4),
                (("Castling black", f"King: {castling['b']['k']}; Queen: {castling['b']['q']}"), 5),
                (("Halfmove clock", f"{self.half_moves_since_pawn()}"), 5),
            ]
            if issubclass(base_ai, MetricProvider):
                ret += base_ai.metrics(self=self)
            return ret

        @property
        def white_name(self):
            return self.name if self._i_am_white else self._enemy_name

        @property
        def black_name(self):
            return self.name if not self._i_am_white else self._enemy_name

        def new_match(self, match_id: UUID, enemy: str, match_format: MatchFormatData):
            self._enemy_name = enemy
            ret = super(_WrappedAI, self).new_match(match_id=match_id, enemy=enemy, match_format=match_format)
            need_update(self)
            return ret

        def finalize_match(self, match_id: UUID, status: MatchStatusData, statistics: str):
            ret = super(_WrappedAI, self).finalize_match(match_id=match_id, status=status, statistics=statistics)
            need_update(self)
            return ret

        def new_game(self, game_id: UUID, match_id: UUID, white_player: str):
            self._i_am_white = white_player == self.name
            ret = super(_WrappedAI, self).new_game(game_id=game_id, match_id=match_id, white_player=white_player)
            need_update(self)
            return ret

        def finalize_game(self, game_id: UUID, match_id: UUID, winner: Optional[str], pgn: str):
            SYSTEM_LOGGER.info(f"Game ended. {winner} ({'you' if winner == self.name else 'not you'}) won")
            ret = super(_WrappedAI, self).finalize_game(game_id=game_id, match_id=match_id, winner=winner, pgn=pgn)
            need_update(self)
            return ret

        def get_move(self, game_id: UUID, match_id: UUID, game_state: GameState) -> MoveData:
            self._fen = game_state.board_state.fen
            self._your_time = game_state.your_time
            self._enemy_time = game_state.enemy_time
            ret = super(_WrappedAI, self).get_move(game_id=game_id, match_id=match_id, game_state=game_state)
            need_update(self)
            return ret

    return _WrappedAI(base_init_values=init_values)



import queue
from typing import List, Tuple, Optional

from asciimatics.exceptions import NextScene, StopApplication
from asciimatics.widgets import (
    Frame, ListBox, Widget, Button, Layout, Divider, MultiColumnListBox, Label, VerticalDivider, TextBox
)
from asciimatics.parsers import AnsiTerminalParser

from j_chess_client_manager.ui.widgets.chessboard import ChessBoard
from j_chess_client_manager.ui.widgets.log import LogList
from j_chess_client_manager.clients import SuperProvider


class MainFrame(Frame):
    def __init__(self, screen, theme: str = "default"):
        super(MainFrame, self).__init__(screen=screen, height=screen.height, width=screen.width,
                                        on_load=self._on_load,
                                        hover_focus=True,
                                        can_scroll=False,
                                        has_border=True,
                                        title="♛ J-Chess Client-Manager 🤖")

        self.set_theme(theme=theme)

        self._ais: List[SuperProvider] = [
            # wrap_ai(base_ai=SampleAI, init_values={"white_name": "Gray"}),
            # wrap_ai(base_ai=SampleAI, init_values={})
        ]

        # Initialize widgets
        self._ai_list = ListBox(
            Widget.FILL_FRAME,
            options=[],
            name="AIs",
            add_scroll_bar=True,
            on_change=self._on_pick,
            on_select=self._edit)
        self._chessboard = ChessBoard()
        self._edit_button = Button("Edit", self._edit)
        self._delete_button = Button("Delete", self._delete)
        self._metrics = MultiColumnListBox(
            Widget.FILL_FRAME,
            columns=["33%", "67%"],
            titles=["Metric", "Value"],
            add_scroll_bar=True,
            options=[],
            parser=AnsiTerminalParser(),
            space_delimiter="|",
        )
        self._metrics._is_tab_stop = False
        self._logs = LogList(name="LogList", height=15)

        # Add widgets to layout
        main_layout = Layout([49, 1, 100, 1, 49], fill_frame=True)
        log_layout = Layout([100], fill_frame=False)
        button_layout = Layout([1, 1, 1, 1], fill_frame=False)
        self.add_layout(main_layout)
        self.add_layout(log_layout)
        self.add_layout(button_layout)

        main_layout.add_widget(Label("Connected clients"), 0)
        main_layout.add_widget(Label("Metrics", align=">"), 4)
        main_layout.add_widget(Divider(), 0)
        main_layout.add_widget(Divider(), 4)
        main_layout.add_widget(self._ai_list, 0)
        main_layout.add_widget(self._chessboard, 2)
        main_layout.add_widget(self._metrics, 4)
        main_layout.add_widget(VerticalDivider(), 1)
        main_layout.add_widget(VerticalDivider(), 3)

        log_layout.add_widget(Divider())
        log_layout.add_widget(self._logs)

        button_layout.add_widget(Divider(), 0)
        button_layout.add_widget(Divider(), 1)
        button_layout.add_widget(Divider(), 2)
        button_layout.add_widget(Divider(), 3)
        button_layout.add_widget(Button("Add", self._add), 0)
        button_layout.add_widget(self._edit_button, 1)
        button_layout.add_widget(self._delete_button, 2)
        button_layout.add_widget(Button("Quit", self._quit), 3)
        self.fix()
        self._on_pick()
        self._set_ais()

    def add_ai(self, new_ai: SuperProvider):
        self._ais.append(new_ai)
        self._set_ais()

    def _set_ais(self):
        def get_name(_x):
            try:
                return _x.name
            except AttributeError:
                return _x.__class__.__name__
        self._ai_list.options = [(f"{x.get_client_type().fixed_represent()} {get_name(x)}", x) for x in self._ais]

    @property
    def current_client(self) -> Optional[SuperProvider]:
        return self._ai_list.value

    def _set_metrics(self, metrics: List[Tuple[Tuple[str, str], int]]):
        self._metrics.options = metrics

    def _on_pick(self):
        val: SuperProvider = self.current_client
        self._edit_button.disabled = True
        self._delete_button.disabled = val is None

        self._set_metrics([] if val is None else val.metrics())
        self._chessboard.data_provider = val

    def _update(self, frame_no):
        super()._update(frame_no)

    def _on_load(self, new_value=None):
        pass

    def _add(self):
        raise NextScene("Add Client")

    def _edit(self):
        pass

    def _delete(self):
        pass

    @staticmethod
    def _quit():
        raise StopApplication("User pressed quit")



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

try:
    with open('HISTORY.rst') as history_file:
        history = history_file.read()
except FileNotFoundError:
    history = ""

requirements = ["j-chess-lib", "asciimatics"]

test_requirements = [ ]

setup(
    author="RedRem95",
    author_email='redrem@botschmot.de',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="Manager for clients designed for the awesome j-chess-server using the j-chess-lib for python",
    entry_points={
        'console_scripts': [
            'j_chess_client_manager=j_chess_client_manager.__main__:main',
        ],
    },
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme + ('\n\n' if len(history) > 0 else '') + history,
    include_package_data=True,
    keywords='j_chess_client_manager',
    name='j_chess_client_manager',
    packages=find_packages(include=['j_chess_client_manager', 'j_chess_client_manager.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/RedRem95/j_chess_client_manager',
    version='0.1.5',
    zip_safe=False,
)



from enum import Enum
import inspect
from typing import Dict, Callable, Any

from asciimatics.exceptions import NextScene, ResizeScreenError, StopApplication, InvalidFields, Highlander
from asciimatics.screen import Screen
from asciimatics.widgets import (
    Frame, Layout, Text, TextBox, Button, DropdownList, RadioButtons, Label, Divider, PopUpDialog
)
from j_chess_lib.ai import AI, StoreAI, DumbAI
from j_chess_lib.ai.Sample import SampleAI
from j_chess_lib.client import Client
from j_chess_lib.communication import Connection

from j_chess_client_manager.ui.utilities import get_widget_by_parameter, get_all_ais
from j_chess_client_manager.clients.ai_wrapper import wrap_ai
from j_chess_client_manager.clients import SuperProvider
from j_chess_client_manager.ui.widgets.log import LogList
from j_chess_client_manager.logging import SYSTEM_LOGGER


class ClientTypes(Enum):
    AI = 0
    Spectator = 1


IP_VALIDATOR = r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$"
PORT_VALIDATOR = r"^([0-9]{1,4}|[1-5][0-9]{4}|6[0-4][0-9]{3}|65[0-4][0-9]{2}|655[0-2][0-9]|6553[0-5])$"


class ClientView(Frame):

    def __init__(self, screen, ai_adder: Callable[[SuperProvider], None], theme: str = "green"):
        height = min(max(screen.height * 2 // 3, 30), screen.height)
        width = min(max(screen.width * 2 // 3, 100), screen.width)
        super(ClientView, self).__init__(screen, height, width, hover_focus=True, title=f"Add new client +🤖")
        self._ai_adder = ai_adder
        self._convertors: Dict[str, Callable[[Any], Any]] = {}

        self.set_theme(theme=theme)

        self._type_selector = RadioButtons(options=[(x.name, x) for x in sorted(ClientTypes, key=lambda x: x.value)],
                                           label="Select the type of client to create",
                                           name="TypeSelector",
                                           on_change=self._select_type)
        self._ok_button = Button("OK", self._ok)
        self._logs = LogList(name="LogList", height=15)

        # Create the form for displaying the list of contacts.
        selection_layout = Layout([100], fill_frame=False)
        self._type_layout = Layout([100], fill_frame=False)
        self._component_layout = Layout([100], fill_frame=True)
        self.add_layout(selection_layout)
        self.add_layout(self._type_layout)
        self.add_layout(self._component_layout)
        selection_layout.add_widget(self._type_selector)

        message_layout = Layout([100], fill_frame=False)
        bottom_layout = Layout([1, 1, 1, 1], fill_frame=False)
        self.add_layout(message_layout)
        self.add_layout(bottom_layout)
        message_layout.add_widget(Label("Use \"<<None>>\" if you want to pass none in any text field", align="^"), 0)
        message_layout.add_widget(Divider())
        message_layout.add_widget(self._logs)
        for _ in range(4):
            bottom_layout.add_widget(Divider(), _)
        bottom_layout.add_widget(self._ok_button, 0)
        bottom_layout.add_widget(Button("Cancel", self._cancel), 3)

        self.fix()

    def _build_ai_ui(self):
        self._ok_button.disabled = True

        self._type_layout.add_widget(Divider())

        for param in inspect.signature(Connection).parameters.values():
            widget, widget_converter = get_widget_by_parameter(name_base="Connection", parameter=param)
            self._convertors[widget.name] = widget_converter
            self._type_layout.add_widget(widget)

        self._type_layout.add_widget(Divider())

        available_ais = get_all_ais()
        # SYSTEM_LOGGER.info(f"Found {len(available_ais)} AI classes")

        if len(available_ais) > 0:

            ai_selector = [None]

            def setup_components():
                self._component_layout.clear_widgets()

                selected_ai = ai_selector[0].value
                if selected_ai is None:
                    self._component_layout.add_widget(Label("Please select your AI-Class above"))
                    self._ok_button.disabled = True
                else:
                    ai_parameters = inspect.signature(selected_ai).parameters.values()
                    if len(ai_parameters) > 0:
                        for ai_param in ai_parameters:
                            ai_widget, ai_widget_converter = get_widget_by_parameter(name_base=selected_ai.__name__,
                                                                                     parameter=ai_param)
                            self._convertors[ai_widget.name] = ai_widget_converter

                            self._component_layout.add_widget(ai_widget)
                    else:
                        self._component_layout.add_widget(Label("It seems like your AI does not want any parameters"
                                                                ". You are good to go. :)"))
                    self._ok_button.disabled = False

                self.fix()

            ai_selector[0] = DropdownList([("Select Here", None)] + [(x.__name__, x) for x in available_ais],
                                          label="Select your AI", name="AI__class", on_change=setup_components)
            setup_components()

            # noinspection PyTypeChecker
            self._type_layout.add_widget(ai_selector[0])
            self._type_layout.add_widget(Divider())
        else:
            self._type_layout.add_widget(Label("You have to provide at least one implemented AI-class.\n"
                                               "Maybe you forgot to include the package its in? If so"
                                               "use --with-package to point to it"))

    def _build_spectator_ui(self):
        self._type_layout.add_widget(Label("Spectators are not yet supported... :("))
        self._ok_button.disabled = True

    def _select_type(self):
        self._type_layout.clear_widgets()
        self._component_layout.clear_widgets()
        t = self._type_selector.value
        if t == ClientTypes.AI:
            self._build_ai_ui()
        elif t == ClientTypes.Spectator:
            self._build_spectator_ui()

        self.fix()

    def reset(self):
        super(ClientView, self).reset()

    def _build_ai(self):
        from pprint import pformat
        connection_prefix = "Connection__"
        connection_parameters = {
            k[len(connection_prefix):]: self._convertors[k](v) for k, v in self.data.items() if
            str(k).startswith(connection_prefix)
        }
        ai_class = self.data["AI__class"]
        ai_parameter = {
            k[len(ai_class.__name__) + 2:]: self._convertors[k](v) for k, v in self.data.items() if
            str(k).startswith(ai_class.__name__)
        }
        SYSTEM_LOGGER.info(f"Creating connection: {[(k, type(v), v) for k, v in connection_parameters.items()]}")
        connection = Connection(**connection_parameters)

        ai = wrap_ai(ai_class, init_values=ai_parameter, need_update=self._invalidate_frame)
        SYSTEM_LOGGER.info(f"Created connection {connection}; Created AI {ai}")
        client = Client(connection=connection, ai=ai)
        client.start()
        SYSTEM_LOGGER.info(f"Started client {client}")
        self._ai_adder(ai)
        # raise Exception(f"{pformat(connection_parameters)}\n{ai_class}\n{pformat(ai_parameter)}")

    def _invalidate_frame(self, *args, **kwargs):
        screen: Screen = self.screen
        screen.force_update()

    def _ok(self):
        self.save()
        try:

            t = self.data["TypeSelector"]

            if t == ClientTypes.AI:
                self._build_ai()

            raise NextScene("Main")
        except (NextScene, ResizeScreenError, StopApplication, InvalidFields, Highlander) as e:
            raise e
        except Exception as e:
            from pprint import pformat
            self.scene.add_effect(PopUpDialog(
                screen=self.screen,
                text=f"Could not create new client.\n{type(e).__name__}: {e}",
                buttons=["Ok"],
                on_close=None)
            )
            raise e

    def _cancel(self, *args, **kwargs):
        self.reset()
        raise NextScene("Main")



"""Console script for j_chess_client_manager."""
from . import logging
import argparse
import sys
from typing import Any
import importlib

from asciimatics.screen import Screen
from asciimatics.exceptions import ResizeScreenError
from asciimatics.widgets.utilities import THEMES

from .ui import run_function_creator


def main():

    available_themes = list(THEMES.keys())
    default_theme = available_themes[0]

    parser = argparse.ArgumentParser(prog="j_chess_client_manager",
                                     description="Program to manage your clients and see their playstate")

    parser.add_argument("--theme", dest="theme", choices=available_themes, required=False, default=default_theme,
                        help=f"Set the theme used for the application [Default: \"{default_theme}\"]")
    parser.add_argument("--with-package", dest="package", type=str, nargs="+", required=False, default=tuple(),
                        help="Set package to be included. Should point to a package that then imports your AI so it can"
                             "be detected")

    args = parser.parse_args()

    for package in args.package:
        package: str
        try:
            importlib.import_module(package)
            print(f"Imported \"{package}\"")
        except ModuleNotFoundError:
            print(f"Package \"{package}\" could not be imported")

    last_scene: Any = None

    while True:
        try:
            Screen.wrapper(run_function_creator(start_scene=last_scene, theme=args.theme),
                           catch_interrupt=False, arguments=[])
            return 0
        except KeyboardInterrupt:
            return 1
        except ResizeScreenError as e:
            last_scene = e.scene
            pass

    return 0


if __name__ == "__main__":
    sys.exit(main())
